#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ext/minIni_07/minIni.h"
#include "log.h"
#include "event_config.h"


#define INI_NA ((char*) "INI N/A")
#define ini_is_na(str) (((char*) str != (char*) NULL) && (strcmp(str, INI_NA) == 0))
#define INI_BUF_SIZE 2048

static inline int __safe_ini_gets(const char *section, const char *key, const char *def,
        char *buffer, size_t buffer_size, const char *filename) {
    ini_gets(section, key, def, buffer, buffer_size, filename);
    if (errno) {
        log_error("Error reading from event file %s: %s", filename, strerror(errno));
        return errno;
    }
    return 0;
}

extern event_config_t* scan_event_file (const char *filename) {
    char ini_section[INI_BUF_SIZE], ini_value[INI_BUF_SIZE], ini_key[INI_BUF_SIZE];
    int version, i, s;

    if (__safe_ini_gets(NULL, "version", INI_NA, ini_value, INI_BUF_SIZE, filename))
        return NULL;
    version = atoi(ini_value);
    if (ini_is_na(ini_value)) {
        log_error("Could not find version information in file %s", filename);
        return NULL;
    } else if (version < MIN_VERSION || version > MAX_VERSION) {
        log_error("Illegal version value: %s; must be between %d and %d",
                ini_value, MIN_VERSION, MAX_VERSION);
        return NULL;
    }


    event_config_t *event_config = calloc(1, sizeof(event_config_t));
    event_config->version = version;
    if (1 == version) {
        log_debug("%s is version 1 file, handling accordingly", filename);
        event_config->listener_count = 1;
        event_config->listeners = calloc(event_config->listener_count, sizeof(event_listener_t*));
        event_listener_t *listener = calloc(1, sizeof(event_listener_t));
        event_config->listeners[0] = listener;
        listener->type = DIRECTORY_CHANGE;
        listener->name = strdup("default version 1 listener");
        listener->cwd = false;
        listener->media_type = VIDEO;
        for (i = 0; ini_getkey(NULL, i, ini_key, INI_BUF_SIZE, filename) > 0; i++) {
            if (__safe_ini_gets(NULL, ini_key, INI_NA, ini_value, INI_BUF_SIZE, filename)) {
                // TODO: free elements of listener
                free(listener);
                free(event_config->listeners);
                free(event_config);
                return NULL;
            }
            if (strcmp("pattern", ini_key) == 0) {
                listener->pattern = strdup(ini_value);
            } else if (strcmp("script", ini_key) == 0) {
                listener->script = strdup(ini_value);
            } else {
                log_info("ignoring value for %s", ini_key);
            }
        }
        log_debug("%s: pattern=%s, script=%s", filename, listener->pattern, listener->script);
    } else if (2 == version) {
        log_debug("%s is version 2 file, handling accordingly", filename);
        for (s = 0; ini_getsection(s, ini_section, INI_BUF_SIZE, filename) > 0; s++) {
            event_listener_t *listener = calloc(1, sizeof(event_listener_t));
            event_config->listeners = realloc(event_config->listeners,
                    (event_config->listener_count + 1) * sizeof(event_listener_t*));
            event_config->listeners[event_config->listener_count] = listener;
            event_config->listener_count++;

            listener->name = strdup(ini_section);
            listener->type = DIRECTORY_CHANGE;
            for (i = 0; ini_getkey(ini_section, i, ini_key, INI_BUF_SIZE, filename) > 0; i++) {
                if (__safe_ini_gets(ini_section, ini_key, INI_NA, ini_value, INI_BUF_SIZE, filename)) {
                    for (i = 0; i < event_config->listener_count; i++) {
                        // TODO: free elements of listener
                        free(event_config->listeners[i]);
                    }
                    free(event_config->listeners);
                    free(event_config);
                    return NULL;
                }
                if (strcmp("type", ini_key) == 0) {
                    if (strcmp(LISTENER_TYPE_DIRECTORY_CHANGE, ini_value) == 0) {
                        listener->type = DIRECTORY_CHANGE;
                    } else {
                        log_warn("%s is an illegal value for %s, ignoring", ini_value, ini_key);
                    }
                } else if (strcmp("pattern", ini_key) == 0) {
                    listener->pattern = strdup(ini_value);
                } else if (strcmp("script", ini_key) == 0) {
                    listener->script = strdup(ini_value);
                } else if (strcmp("cwd", ini_key) == 0) {
                    if (strcmp("true", ini_value) == 0) {
                        listener->cwd = true;
                    } else if (strcmp("false", ini_value) == 0) {
                        listener->cwd = false;
                    } else {
                        log_warn("%s is an illegal value for %s, ignoring", ini_value, ini_key);
                    }
                } else if (strcmp("media_type", ini_key) == 0) {
                    if (strcmp(MEDIA_TYPE_VIDEO, ini_value) == 0) {
                        listener->media_type = VIDEO;
                    } else if (strcmp(MEDIA_TYPE_AUDIO, ini_value) == 0) {
                        listener->media_type = AUDIO;
                    } else if (strcmp(MEDIA_TYPE_PHOTO, ini_value) == 0) {
                        listener->media_type = PHOTO;
                    } else if (strcmp(MEDIA_TYPE_ALL, ini_value) == 0) {
                        listener->media_type = ALL;
                    } else {
                        log_warn("%s is an illegal value for %s, ignoring", ini_value, ini_key);
                    }
                } else {
                    log_info("ignoring value for %s", ini_key);
                }
            }
            log_debug("%s: type=%d, name=%s, pattern=%s, script=%s, cwd=%s, media_type=%d",
                    filename, listener->type, listener->name, listener->pattern, listener->script,
                    listener->cwd ? "true" : "false", (int) listener->media_type);
        }
    }

    return event_config;
}


extern event_config_t** scan_event_files (const event_files_t *event_files) {
    if (NULL == event_files) {
        return NULL;
    }
    int i;
    event_config_t **event_configs = malloc(sizeof(event_config_t*) * event_files->count);
    for (i = 0; i < event_files->count; i++) {
        event_config_t *config = scan_event_file(event_files->files[i]);
        if (config == NULL) {
            log_error("could not parse event file %s", event_files->files[i]);
            return NULL;
        }
        event_configs[i] = config;
    }

    return event_configs;
}

