/**
 * A simple filesystem used to monitor events in the /osd directory of the
 * WDTV.
 *
 *  @author Elmar Weber <wdtv@elmarweber.org>
 *
 *
 * Source based on fusexmp.c from FUSE
 *
 *     FUSE: Filesystem in Userspace
 *     Copyright (C) 2001-2007  Miklos Szeredi <miklos@szeredi.hu>
 *
 *     This program can be distributed under the terms of the GNU GPL.
 *     See the file COPYING.
 *
 */

#define FUSE_USE_VERSION 26


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#endif


#include <stdlib.h>
#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/time.h>
#include <stdbool.h>
#ifdef HAVE_SETXATTR
#include <sys/xattr.h>
#endif


#include "log.h"
#include "memory.h"
#include "find_location.h"
#include "options.h"
#include "event_config.h"
#include "ext/md5.h"

static int log_i = 0;
static event_files_t event_files;
static event_config_t **event_configs;


#define log_function() log_debug("%10d %5d %20s %s", log_i, fuse_get_context()->pid, __FUNCTION__, path); \
        log_i++;


#define log_ft_function() log_debug("%10d %5d %20s %s > %s", log_i, fuse_get_context()->pid, __FUNCTION__, from, to); \
        log_i++;


static int get_config_checksum (unsigned char checksum[16]) {
    int result = md5_file("/conf/config", checksum);
    if (result) {
        log_error("could not update config checksum: error %s %s: %s",
                result == 1 ? "opening" : "reading", "/conf/config",
                strerror(errno));
        return 1;
    } else {
        return 0;
    }
}

#define BASE16SYM			("0123456789ABCDEF")
#define BASE16VAL			("\x0\x1\x2\x3\x4\x5\x6\x7\x8\x9|||||||\xA\xB\xC\xD\xE\xF")

#define BASE16_ENCODELO(b)		(BASE16SYM[((unsigned char)(b)) >> 4])
#define BASE16_ENCODEHI(b)		(BASE16SYM[((unsigned char)(b)) & 0xF])

static inline void __to_hex (char *hex, const unsigned char checksum[16]) {
    int i, j;
    for (i = 0, j = 0; i < 16; i++, j+=2) {
        hex[j] = BASE16_ENCODELO(checksum[i]);
        hex[j + 1] = BASE16_ENCODEHI(checksum[i]);
    }
    hex[j] = '\0';
}


static int get_dmaosd_pid () {
    char cmdout[1024];
    FILE *file = popen("pidof -s dmaosd", "r");
    if (NULL == file) {
	log_error("Could not start pidof program: %s", strerror(errno));
	return -1;
    }
    memset(cmdout, 0, 1024);
    int read = fread(cmdout, 1024, 1, file);
    if (read < 0) {
	log_error("could not read from pidof program: %s", strerror(errno));
	return -1;
    }
    pclose(file);
    int pid = atoi(cmdout);
    if (0 == pid) {
        log_error("could not determine pid based on output \"%s\"", cmdout);
        return -1;
    }

    log_debug("determined pid of dmaosd as %d", pid);

    return pid;
}


static bool loc_addr_determined = false;
static bool loc_addr_determined_list = false;
static unsigned char config_checksum[16];
static int dmaosd_pid = 0;
static unsigned int loc_offset = 0;
static unsigned int loc_offset_thumb = 0;
static unsigned int loc_offset_list = 0;

static void handle_getattr_event (const char *path) {
    if (strstr(path, "basic_browse_thumb") == NULL
    	    && strstr(path, "basic_browse.xml") == NULL)
	return;

    int i, j;

    /* check if the configuration and so possible memory addresses changed,
     * this happens for example when the browsing mode was changed */
    unsigned char current_config_checksum[16];
    if (get_config_checksum(current_config_checksum)) {
        log_error("could not determine directory");
        return;
    }
    char checksum_str[33];
    __to_hex(checksum_str, current_config_checksum);
    log_debug("config checksum: %s", checksum_str);
    if (memcmp(current_config_checksum, config_checksum, 16) != 0) {
	loc_addr_determined = false;
	loc_addr_determined_list = false;
	memcpy(config_checksum, current_config_checksum, 16);
	log_debug("config changed, updating memory config");
    }

    /* update memory addresses for thumb mode*/
    if (!loc_addr_determined && strstr(path, "basic_browse_thumb") != NULL) {
	if ((dmaosd_pid = get_dmaosd_pid()) <= 0) {
	    log_error("Could not determine dmaosd pid");
	    return;
	}
	if (update_memory_addresses(dmaosd_pid)) {
	    log_error("Could not determine dmaosd memory addresses");
	    return;
	}
	if ((loc_offset_thumb =
	     find_location_offset(dmaosd_pid, get_heap_start(), get_heap_end(),
				  get_stack_start(), get_stack_end())) <= 0) {
	    log_error("Could not find location address");
	    return;
	}
	loc_addr_determined = true;
	log_debug("using location address offset of %d (%#x) bytes", loc_offset_thumb, loc_offset_thumb);
    }
    /* update memory addresses for list mode*/
    if (!loc_addr_determined_list && strstr(path, "basic_browse.xml") != NULL) {
	if ((dmaosd_pid = get_dmaosd_pid()) <= 0) {
	    log_error("Could not determine dmaosd pid");
	    return;
	}
	if (update_memory_addresses(dmaosd_pid)) {
	    log_error("Could not determine dmaosd memory addresses");
	    return;
	}
	if ((loc_offset_list =
	     find_location_offset(dmaosd_pid, get_heap_start(), get_heap_end(),
				  get_stack_start(), get_stack_end())) <= 0) {
	    log_error("Could not find location address");
	    return;
	}
	loc_addr_determined_list = true;
	log_debug("using location address offset of %d (%#x) bytes", loc_offset_list, loc_offset_list);
    }
    loc_offset = strstr(path, "basic_browse.xml") == NULL ? loc_offset_thumb : loc_offset_list ;

    char cmd[1024], cmdout[1024], *location;
    sprintf(cmd, "/apps/event-osd/bin/get_path %u %u %d", get_stack_start(),
	    loc_offset, dmaosd_pid);
    FILE *file = popen(cmd, "r");
    if (NULL == file) {
	log_error("Could not start get_path program: %s", strerror(errno));
	return;
    }
    memset(cmdout, 0, 1024);
    int read = fread(cmdout, 1024, 1, file);
    if (read < 0) {
	log_error("could not read from get_path program: %s",
		  strerror(errno));
	return;
    }
    pclose(file);
    location = cmdout;

    log_info("Navigated to %s", location);
    
    char mode[6];

    /* notify registered applications */
    for (i = 0; i < event_files.count; i++) {
        event_config_t *event_config = event_configs[i];
        for (j = 0; j < event_config->listener_count; j++) {
            event_listener_t *listener = event_config->listeners[j];
            if (atoi(&location[0]) == (int) listener->media_type || (int) listener->media_type == 0 ) {
		if (strstr(location, listener->pattern) != NULL) {
                	if (listener->cwd) {
                	    if (chdir(location+1)) {
                	        log_error("could not change working directory to %s: %s",
                	                location+1, strerror(errno));
                	    } else {
                	        log_debug("changed working directory to %s", location+1);
                	    }
                	}
			
			switch (location[0]) {
				case '1' :
					sprintf(mode,"audio");
					break;
				case '2' :
					sprintf(mode,"photo");
					break;
				case '3' :
					sprintf(mode,"video");
					break;
			}

                	sprintf(cmd, "%s \"%s\" \"%s\" \"%s\"",
                	        listener->script,
                	        strstr(path, "basic_browse.xml") == NULL ? "thumb" : "list",
                	        location+1,
				mode );
                	log_debug("executing: %s", cmd);
                	int result = system(cmd);
                	log_info("executed script with result %d: %s", result, cmd);

                	if (listener->cwd) {
                	    if (chdir("/")) {
                	        log_error("could not change working directory to %s: %s",
                	                "/", strerror(errno));
                	    } else {
                	        log_debug("changed working directory to %s", "/");
                	    }
                	}
            	}
            }
        }
    }

    return;
}


static int eosd_getattr (const char *path, struct stat *stbuf) {
//    log_function();
    handle_getattr_event(path);


    int res;

    res = lstat(path, stbuf);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_access (const char *path, int mask) {
    log_function();

    int res;

    res = access(path, mask);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_readlink (const char *path, char *buf, size_t size) {
//    log_function();

    int res;

    res = readlink(path, buf, size - 1);
    if (res == -1)
	return -errno;

    buf[res] = '\0';
    return 0;
}


static int eosd_readdir (const char *path, void *buf, fuse_fill_dir_t filler,
	     off_t offset, struct fuse_file_info *fi) {
//    log_function();

    DIR *dp;
    struct dirent *de;

    (void) offset;
    (void) fi;

    dp = opendir(path);
    if (dp == NULL)
	return -errno;

    while ((de = readdir(dp)) != NULL) {
	struct stat st;
	memset(&st, 0, sizeof(st));
	st.st_ino = de->d_ino;
	st.st_mode = de->d_type << 12;
	if (filler(buf, de->d_name, &st, 0))
	    break;
    }

    closedir(dp);
    return 0;
}

static int eosd_mknod (const char *path, mode_t mode, dev_t rdev) {
//    log_function();

    int res;

    /* On Linux this could just be 'mknod(path, mode, rdev)' but this
       is more portable */
    if (S_ISREG(mode)) {
	res = open(path, O_CREAT | O_EXCL | O_WRONLY, mode);
	if (res >= 0)
	    res = close(res);
    }
    else if (S_ISFIFO(mode))
	res = mkfifo(path, mode);
    else
	res = mknod(path, mode, rdev);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_mkdir (const char *path, mode_t mode) {
//    log_function();

    int res;

    res = mkdir(path, mode);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_unlink (const char *path) {
 //   log_function();

    int res;

    res = unlink(path);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_rmdir (const char *path) {
//    log_function();

    int res;

    res = rmdir(path);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_symlink (const char *from, const char *to) {
//    log_ft_function();

    int res;

    res = symlink(from, to);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_rename (const char *from, const char *to) {
//    log_ft_function();

    int res;

    res = rename(from, to);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_link (const char *from, const char *to) {
//    log_ft_function();

    int res;

    res = link(from, to);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_chmod (const char *path, mode_t mode) {
//    log_function();

    int res;

    res = chmod(path, mode);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_chown (const char *path, uid_t uid, gid_t gid) {
//    log_function();

    int res;

    res = lchown(path, uid, gid);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_truncate (const char *path, off_t size) {
  //  log_function();

    int res;

    res = truncate(path, size);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_utimens (const char *path, const struct timespec ts[2]) {
    //log_function();

    int res;
    struct timeval tv[2];

    tv[0].tv_sec = ts[0].tv_sec;
    tv[0].tv_usec = ts[0].tv_nsec / 1000;
    tv[1].tv_sec = ts[1].tv_sec;
    tv[1].tv_usec = ts[1].tv_nsec / 1000;

    res = utimes(path, tv);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_open (const char *path, struct fuse_file_info *fi) {
//    log_function();

    int res;

    res = open(path, fi->flags);
    if (res == -1)
	return -errno;

    close(res);
    return 0;
}

static int eosd_read (const char *path, char *buf, size_t size, off_t offset,
	  struct fuse_file_info *fi) {
//    log_function();

    int fd;
    int res;

    (void) fi;
    fd = open(path, O_RDONLY);
    if (fd == -1)
	return -errno;

    res = pread(fd, buf, size, offset);
    if (res == -1)
	res = -errno;

    close(fd);
    return res;
}

static int eosd_write (const char *path, const char *buf, size_t size,
	   off_t offset, struct fuse_file_info *fi) {
//    log_function();

    int fd;
    int res;

    (void) fi;
    fd = open(path, O_WRONLY);
    if (fd == -1)
	return -errno;

    res = pwrite(fd, buf, size, offset);
    if (res == -1)
	res = -errno;

    close(fd);
    return res;
}

static int eosd_statfs (const char *path, struct statvfs *stbuf) {
//    log_function();

    int res;

    res = statvfs(path, stbuf);
    if (res == -1)
	return -errno;

    return 0;
}

static int eosd_release (const char *path, struct fuse_file_info *fi) {
//    log_function();


    /* Just a stub.  This method is optional and can safely be left
       unimplemented */

    (void) path;
    (void) fi;
    return 0;
}

static int eosd_fsync (const char *path, int isdatasync, struct fuse_file_info *fi) {
 //   log_function();

    /* Just a stub.  This method is optional and can safely be left
       unimplemented */

    (void) path;
    (void) isdatasync;
    (void) fi;
    return 0;
}

#ifdef HAVE_SETXATTR
/* xattr operations are optional and can safely be left unimplemented */
static int
eosd_setxattr(const char *path, const char *name, const char *value,
	      size_t size, int flags)
{
    log_function();

    int res = lsetxattr(path, name, value, size, flags);
    if (res == -1)
	return -errno;
    return 0;
}

static int
eosd_getxattr(const char *path, const char *name, char *value, size_t size)
{
    log_function();

    int res = lgetxattr(path, name, value, size);
    if (res == -1)
	return -errno;
    return res;
}

static int
eosd_listxattr(const char *path, char *list, size_t size)
{
    log_function();

    int res = llistxattr(path, list, size);
    if (res == -1)
	return -errno;
    return res;
}

static int
eosd_removexattr(const char *path, const char *name)
{
    log_function();

    int res = lremovexattr(path, name);
    if (res == -1)
	return -errno;
    return 0;
}
#endif /* HAVE_SETXATTR */

static struct fuse_operations eosd_operations = {
    .getattr = eosd_getattr,
    .access = eosd_access,
    .readlink = eosd_readlink,
    .readdir = eosd_readdir,
    .mknod = eosd_mknod,
    .mkdir = eosd_mkdir,
    .symlink = eosd_symlink,
    .unlink = eosd_unlink,
    .rmdir = eosd_rmdir,
    .rename = eosd_rename,
    .link = eosd_link,
    .chmod = eosd_chmod,
    .chown = eosd_chown,
    .truncate = eosd_truncate,
    .utimens = eosd_utimens,
    .open = eosd_open,
    .read = eosd_read,
    .write = eosd_write,
    .statfs = eosd_statfs,
    .release = eosd_release,
    .fsync = eosd_fsync,
#ifdef HAVE_SETXATTR
    .setxattr = eosd_setxattr,
    .getxattr = eosd_getxattr,
    .listxattr = eosd_listxattr,
    .removexattr = eosd_removexattr,
#endif
};



int main (int argc, char *argv[]) {
    struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
    if (-1 == parse_options(&args, &event_files)) {
        fprintf(stderr, "Could not parse command line arguments\n");
        return 1;
    }

    if (log_init("/apps/event-osd/event_osd_fs.log") != 0) {
        fprintf(stderr, "Could not open log file %s: %s\n",
                "/apps/event-osd/event_osd_fs.log", strerror(errno));
        return 1;
    }

    if (NULL == (event_configs = scan_event_files(&event_files))) {
        fprintf(stderr, "Could not parse event files\n");
        return 1;
    }


    memset(config_checksum, 0, 16);

    int result = fuse_main(args.argc, args.argv, &eosd_operations, NULL);

    fuse_opt_free_args(&args);

    return result;
}
