/*
 * This file is part of WDTV Tools (http://wdtvtools.sourceforge.net/).
 *
 * Copyright (C) 2009 Elmar Weber <wdtv@elmarweber.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/** 
 * A memory analyzer that searches for a specific string in the heap and then
 * finds all adresses where it is referenced. The reference lookup is done
 * recursive until no more references are found.
 *
 * @author Elmar Weber <wdtv@elmarweber.org>
 * 
 * Based on some code from Pierre Betouin <pierre.betouin@security-labs.org>
 */


#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ptrace.h>
#include <stdlib.h>
#include <string.h>

#include "log.h"


#define attach() ptrace(PTRACE_ATTACH, pid, 0, 0); \
        log_debug("ptrace_attach %d", pid);

#define detach() ptrace(PTRACE_DETACH, pid, 0, 0); \
        log_debug("ptrace_detach %d", pid);

static unsigned int find_reference (char *heap, size_t heap_length, char *stack,
	       size_t stack_length, unsigned int match_addr, 
	       unsigned int heap_start, unsigned int stack_start) {
    unsigned int j;
    unsigned int found_addr = 0;
    unsigned int found_count = 0;
    for (j = 0; j < stack_length - sizeof(match_addr); j++) {
	if (memcmp(stack + j, &match_addr, sizeof(match_addr)) == 0) {
	    log_debug("    found stack reference for %#x at %#x (%#x)",
		      match_addr, (unsigned int) (stack_start + j), j);
            if (0 == found_addr) {
                found_addr = j;
            }
            found_count++;
	}
    }
    log_debug("  found %d occurences", found_count);
    return found_count == 4 ? found_addr : 0;
}


extern unsigned int find_location_offset (int pid, unsigned int heap_start, unsigned int heap_end,
		     unsigned int stack_start, unsigned int stack_end) {
    FILE *mem = NULL;
    char p[64];
    unsigned long i = 0;

    snprintf(p, 63, "/proc/%d/mem", pid);
    attach();

    if (!(mem = fopen(p, "r"))) {
	detach();
	log_error("Could not open memory file for pid %d: %s", pid,
		  strerror(errno));
	return 0;
    }

    if (!fseek(mem, heap_start, SEEK_SET) < 0) {
	log_error("Could not seek file for pid %d: %s", pid, strerror(errno));
	fclose(mem);
	detach();
	return 0;
    }

    size_t heap_length = heap_end - heap_start;
    char *heap = malloc(heap_length);
    if (1 != fread(heap, heap_length, 1, mem)) {
	log_error("Could not read heap memory for pid %d: %s", pid,
		  strerror(errno));
	fclose(mem);
	detach();
	free(heap);
	return 0;
    }

    if (!fseek(mem, stack_start, SEEK_SET) < 0) {
	log_error("Could not seek file for pid %d: %s", pid, strerror(errno));
	fclose(mem);
	detach();
	free(heap);
	return 0;
    }

    size_t stack_length = stack_end - stack_start;
    char *stack = malloc(stack_length);
    if (1 != fread(stack, stack_length, 1, mem)) {
	detach();
	log_error("Could not read heap memory for pid %d: %s", pid,
		  strerror(errno));
	fclose(mem);
	detach();
	free(stack);
	free(heap);
	return 0;
    }

    fclose(mem);
    detach();

    char *actual_match = strdup("/tmp/media/usb");
    int match_len = strlen(actual_match);
    char *match = malloc(match_len + 10);
    memcpy(match + 1, actual_match, match_len + 1);
    match[0] = 0x31;		// movies
    char *match2 = malloc(match_len + 10);
    memcpy(match2 + 1, actual_match, match_len + 1);
    match2[0] = 0x32;		// movies
    char *match3 = malloc(match_len + 10);
    memcpy(match3 + 1, actual_match, match_len + 1);
    match3[0] = 0x33;		// movies
    match_len++;
    unsigned int location_addr_offset = 0;
    for (i = 0; i < heap_length - match_len; i++) {
	if (memcmp(heap + i, match, match_len) == 0 || memcmp(heap + i, match2, match_len) == 0 || memcmp(heap + i, match3, match_len) == 0 ) {
            unsigned int match_addr = (unsigned int) (heap_start + i);
	    log_debug("found at %#x: %s", match_addr, heap + i);
	    unsigned int possible_offset;
            if ((possible_offset =
                    find_reference(heap, heap_length, stack, stack_length,
                        match_addr, heap_start, stack_start)) != 0) {
                location_addr_offset = possible_offset;
		break;
            }
	}
    }

    free(stack);
    free(heap);
    free(match);
    free(match2);
    free(match3);
    free(actual_match);

    return location_addr_offset;
}
