/**
 * Contains functions to parse the command line options via the FUSE options library
 * and functions to manage the global configuration.
 *
 * @author Elmar Weber <wdtv@elmarweber.org>
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "options.h"


/**
 * Enumeration for options that require processing in <code>eosd_opt_proc</code>.
 */
enum {
     KEY_HELP,
     KEY_VERSION,
     KEY_EVENT_FILE,
};


static struct fuse_opt eosd_opts[] = {
    FUSE_OPT_KEY("-V",        KEY_VERSION),
    FUSE_OPT_KEY("--version", KEY_VERSION),
    FUSE_OPT_KEY("-h",        KEY_HELP),
    FUSE_OPT_KEY("--help",    KEY_HELP),
    FUSE_OPT_KEY("-e %s",     KEY_EVENT_FILE),

    FUSE_OPT_END
};

#define MAX_EVENT_FILES 20

static event_files_t *_event_files;

/*
 * Processes the command line options. Mainly the help and version options. In case
 * the help or version option is passed in the argument list, the information is
 * printed and then the program exits.
 *
 * @param *data FUSE options.
 * @param *arg the argument.
 * @param key the parameter key index as specified via the fuse_opt structure.
 * @param *outargs the arguments passed to fuse_main at a later point.
 *
 * @return 0 in case everything works.
 */
static int eosd_opt_proc(void *data, const char *arg, int key, struct fuse_args *outargs) {
    if(KEY_HELP == key) {
        fprintf(stderr,
                "usage: event-osd [options] moutpoint\n"
                "\n"
                "general options:\n"
                "    -o opt,[opt...]  mount options\n"
                "    -h   --help      print help\n"
                "    -V   --version   print version\n"
                "\n"
                "event-osd options:\n"
                "    -e path          add event configuration file to process\n"
                "                     can be used several times\n"
                );
        fuse_opt_add_arg(outargs, "-ho");
    } else if (KEY_VERSION == key) {
        fprintf(stderr, "event-osd version %s\n", PACKAGE_VERSION);
        fuse_opt_add_arg(outargs, "--version");
    } else if (KEY_EVENT_FILE == key) {
        char *file = malloc(strlen(arg));
        memcpy(file, arg + 2, strlen(arg) - 2);
        if (_event_files->count >= _event_files->capacity) {
            fprintf(stderr, "cannot add events file %s, maximum number of %d files reached",
                    file, _event_files->capacity);
            free(file);
            return 1;
        }
        _event_files->files[_event_files->count] = file;
        _event_files->count++;
    } else {
        return 1;
    }

    return 0;
}


/**
 * Parses the command line options via FUSE.
 *
 * @param *args the FUSE arguments structure.
 * @param *event_files pointer to the event_files_t that is filled with the command line arguments.
 *
 * @return 0 if everything works, -1 if an error occurs while parsing
 *         the command line options.
 */
extern int parse_options (struct fuse_args *args, event_files_t *event_files) {
    if (NULL == event_files) {
        return -1;
    }
    event_files->count = 0;
    event_files->capacity = MAX_EVENT_FILES;
    event_files->files = malloc(sizeof(char*) * MAX_EVENT_FILES);
    _event_files = event_files;
    if (fuse_opt_parse(args, NULL, eosd_opts, eosd_opt_proc) == -1) {
        return -1;
    }

    return 0;
}
